/**
 * P4
 * http://snarfed.org/space/p4
 *
 * Copyright 2001-2004
 * Ken Ashcraft <kashcraft+p4@gmail.com>
 * Ryan Barrett <p4@ryanb.org>
 * Maulik Shah <p4@maulik.net>
 * Nathan Stoll <nathanstoll+p4@gmail.com>
 *
 * @file servicer.cpp
 *
 * The main() for the backend.
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA 02111-1307 USA
 */

#include "conn_mgr.h"
#include "receiveq.h"
#include "msg_data.h"
#include "id_table.h"
#include "servicer.h"
#include "debug.h"
#include "settings.h"
#include "p4error.h"

#ifdef _WIN32           // for gettimeofday()
#include <winsock.h>
#include <sys/timeb.h>
#else
#include <arpa/inet.h>
#include <sys/time.h>
#endif

#include <errno.h>
#include <string>
using namespace std;


/*****************************************************************************
 * GLOBALS
 *****************************************************************************/
string glocal_user = csettings::instance().get<string>("local_username");
pthread_t servicer;  // the main thread for the backend
extern bool done_bootstrapping;


/** servicer_loop
 *
 * The main servicer loop, the heart of the backend. Repeatedly receives on
 * connections, manages connections, purges the id table, etc.
 */
void *servicer_loop(void *data)
{
  cconn_mgr &conn_mgr = cconn_mgr::instance();
  cid_table &id_table = cid_table::instance();

  stop_info *stop = reinterpret_cast<stop_info*>(data);

  // 
  // the main servicer loop is here! GET ME A FUCKING TOWEL!!!
  // 
  while(true) {
    uint ret = conn_mgr.receive();
    if (ret != kerr_success) {
      cdebug::instance() << acquire("servicer_debug")
                         << "receive() failed: " << ret
                         << release();
    }
    if (done_bootstrapping)
      conn_mgr.manage();

    id_table.purge();

    // see if it is time to quit
    if (stop->time_to_stop)
      break;
  }
  cdebug::instance() << acquire("servicer_debug")
                     << "FINISHING LOOP!!!"
                     << release();

  sem_post(&stop->sem);
  return NULL;
}


/********************************************************************
 * TEST MAIN
 *******************************************************************/

#ifdef SERVICER_MAIN

void *talker_loop(void *data)
{
  // just pause for a few seconds
#ifdef _WIN32
  struct timespec now;
#else
  struct timeval now;
#endif
  struct timespec timeout;

  assert(!gettimeofday(&now, NULL));

  timeout.tv_sec = now.tv_sec + 10;
#ifdef _WIN32
  timeout.tv_nsec = now.tv_nsec * 1000;
#else
  timeout.tv_nsec = now.tv_usec * 1000;
#endif
  pthread_cond_t cond = PTHREAD_COND_INITIALIZER;
  pthread_mutex_t mtx = PTHREAD_MUTEX_INITIALIZER;
    
  pthread_mutex_lock(&mtx);
  pthread_cond_timedwait(&cond, &mtx, &timeout);
  pthread_mutex_unlock(&mtx);
  
  pthread_mutex_destroy(&mtx);
  pthread_cond_destroy(&cond);

  assert(!creceiveq::instance().plugin_loaded(1));
  string remote_user = csettings::instance().get<string>("remote_user0");
  string hello = "hello world";
  char from_user[kp4_username_len];
  char *incoming_data = new char[hello.size() + 1];

  while (true) {
    cdebug::instance() << acquire("servicer_debug")
                       << "TALKER: talking to " << remote_user
                       << release();
    
    cmsg_data msg(1, hello.c_str(), hello.size() + 1, false);

    // SEND
    const cconn &conn = msg.send(remote_user.c_str());
    if (conn.get_status() == cconn::efailed) {
      cdebug::instance() << acquire("servicer_debug")
                         << "TALKER: failed to send to " << remote_user
                         << release();
      return NULL;
    }

    // RECEIVE
    int size;
    uint ret = creceiveq::instance().peek(1, size);
    if (ret == kerr_success) {
      if (size > 0) {
        assert(static_cast<uint>(size) == hello.size() + 1);

        int size_written; bool encrypted;
        ret = creceiveq::instance().dequeue(1, from_user, incoming_data, size,
                                            size_written, encrypted);
        if (ret != kerr_success) {
          cdebug::instance() << acquire("servicer_debug")
                             << "TALKER: failed to dequeue "<< ret
                             << release();
          return NULL;
        }
        cdebug::instance() << acquire("servicer_debug")
                           << "TALKER: received \"" << incoming_data << "\""
                           << release();
      }
    } else {
      cdebug::instance() << acquire("servicer_debug")
                         << "TALKER: peek failed"
                         << release();
    }
    cdebug::instance() << acquire("servicer_debug")
                       << "TALKER: looping"
                       << release();
  }
  return NULL;
}


  

int main(int argc, char **argv)
{
  pthread_t servicer, talker;

#ifdef _WIN32
  WSADATA wsaData; // if this doesn't work
  // WSAData wsaData; // then try this instead

    if (WSAStartup(MAKEWORD(1, 1), &wsaData) != 0) {
        fprintf(stderr, "WSAStartup failed.\n");
        exit(1);
    } 
#endif

  sem_t connected, done;
  
  sem_init(&connected, 0, 0);
  sem_init(&done, 0, 0);
  
  uint ret = cconn_mgr::instance().listen();
  if (ret == kerr_success) {
    cdebug::instance() << acquire("servicer_debug") 
                       << "Listening for connections"
                       << release();

    pthread_create(&servicer, NULL, &servicer_loop, (void*)&connected);
  
    pthread_create(&talker, NULL, &talker_loop, (void*)&connected);

    sem_wait(&done);
    return 0;
  }

  cdebug::instance() << acquire("servicer_debug") 
                     << "Could not connect to the network for reason: " << ret
                     << release();
  return -1;
}

#endif  // SERVICER_MAIN
